import itertools
import re
import string
from typing import Iterable, TypeAlias
from collections import defaultdict, deque
from apycula.chipdb import Device, rc2tbrl, rc2tbrl_0
from apycula.gowin_unpack import parse_tile_, tbrl2rc
import random
# from apycula.chipdb_builder import tbrl2rc

#Node Type
Node: TypeAlias = tuple[int, int, str]


def parse_intertile_wire(wire:str):
    # {direction}{length}{number}{segment}
    intertile_regex = r'([NSEW])(\d)(\d)(\d)'
    if m := re.match(intertile_regex, wire):
        return m.groups()
    else:
        return None

def is_intertile_node(wire:str):
    if not wire:
        return False
    if (wire[:2] in ("SN", "EW")):
        return True
    intertile_regex = r'([NSEW])(\d)(\d)(\d)'
    return bool(re.match(intertile_regex, wire))

# Mostly copied from gowin_arch_gen in nextpnr :). Logic for handling cases
# Where a wire would appear to go out of bounds
def uturn(db:Device, node:Node):
    row, col, wire = node
    uturnlut = {'N': 'S', 'S': 'N', 'E': 'W', 'W': 'E'}
    m = re.match(r"([NESW])([128]\d)(\d)", wire)
    if m:
        direction, num, segment = m.groups()
        # wires wrap around the edges
        # assumes 0-based indexes
        if row < 0:
            row = -1 - row
            direction = uturnlut[direction]
        if col < 0:
            col = -1 - col
            direction = uturnlut[direction]
        if row > db.rows - 1:
            row = 2 * db.rows - 1 - row
            direction = uturnlut[direction]
        if col > db.cols - 1:
            col = 2 * db.cols - 1 - col
            direction = uturnlut[direction]
        wire = f'{direction}{num}{segment}'
    return row, col, wire

def source_intertile_wire(db:Device, node:Node):
    """
    Returns the source of an intertile wire or 'None' if the wire supplied is invalid or no source is found
    Args:
        db (Device):
        loc (tuple (x:int, y:int)): coordinates of the tile the node of interest is from
        node: An intertile wire
    Returns:
        source_node (Node): The source Node of an intertile wire
    """

    row, col, wire = node
    loc = (row, col)
    inter_aliases = {
        'E110': 'EW10',
        'W110': 'EW10',
        'E120': 'EW20',
        'W120': 'EW20',
        'S110': 'SN10',
        'N110': 'SN10',
        'S120': 'SN20',
        'N120': 'SN20'
    }
    direction, length, number, segment = parse_intertile_wire(wire)
    uturnlut = {'N': 'S', 'S': 'N', 'E': 'W', 'W': 'E'}
    if not is_intertile_node(wire):
        return None
    reverse_dir = uturnlut[wire[0]]
    vecs = { 'N': (-1,0), 'S': (1,0), 'W': (0,-1), 'E': (0,1) }[reverse_dir]

    trow = row + vecs[0] * int(segment)
    tcol = col + vecs[1] * int(segment)

    srow, scol, rev_wire = uturn(db, (trow, tcol, f'{reverse_dir}{length}{number}{segment}'))
    true_wire = f'{uturnlut[rev_wire[0]]}{length}{number}0'
    if true_wire in inter_aliases:
        return srow, scol, inter_aliases[true_wire]
    return srow, scol, true_wire

_tbrlre = re.compile(r"IO([TBRL])(\d+)(\w)")
def __normalize_pin(db, pin):
    if isinstance(pin, str) and _tbrlre.match(pin):
        row,col,pin_idx = tbrl2rc(db, pin)
        wire = db[row, col].bels["IOB"+pin_idx].portmap["I"]
        return row,col,wire
    else:
        return pin


def next_intertile_wires(db:Device, node:Node):

    """
    Returns the intertile wires that have `node` as their source
    Args:
        db (Device):
        node (Node): A node with an intertile wire

    Returns:
        inter_wires(list[Node]): list of nodes that have `node` as their source
    """

    row, col, wire = node
    if not is_intertile_node(wire):
        return []
    if wire[-1] != '0':
        return []
    inter_wires = []
    if wire[:2]=="SN":
        inter_wires = [uturn(db, (row+disp, col, f'{dir}1{wire[2]}1'))
                        for (disp, dir) in [(-1,"N"), (1, "S")]]

    elif wire[:2]=="EW":
        inter_wires = [uturn(db, (row, col+disp, f'{dir}1{wire[2]}1'))
                        for (disp, dir) in [(1,"E"), (-1, "W")]]

    elif (this_wire:=parse_intertile_wire(wire)):
        direction, length, number, segment = this_wire
        disp_r, disp_c = { 'N': (-1,0), 'S': (1,0), 'W': (0,-1), 'E': (0,1) }[direction]
        next_hops = {1: [1], 2:[1,2], 8:[4,8]}[int(length)]
        inter_wires = [uturn(db, (row+disp_r*hop, col+disp_c*hop, f'{direction}{length}{number}{hop}'))
                    for hop in next_hops]
    return inter_wires



def get_input_path_dict(tile_dict:dict[Node, Node], db:Device, source:Node|str, through_bel=('lut', 'ff')):
    """
    Returns dictionary of connections that emanate from a `node`.
    Args:
        tile_dict (dict): Dictionary of tiles (output of tile_bitmap) .
        db (Device): Device object representing the FPGA or chip.
        source(Node|str): Node to start trace from. The TBRL format may be used to specify IO Nodes
        through_bel (tuple, optional): BEL types to trace through. Defaults to ('lut', 'ff'), the only ones implemented for now.

    Returns:
        dict: A dictionary of dest<-src connections where the keys are destinations.
              In cases where there are multiple conceptual inputs to the same destination (like tracing from the output to input of a bel),
              the input wires are separated by a "#". dest and src are always strings
    """

    LUT_IN_REGEX = r'[A-D][0-7]'
    FF_IN_REGEX = r'F[0-7]'

    source = __normalize_pin(db, source)

    path_dict = {}
    row, col, snode = source
    sloc = (row,col)
    next_tile = deque([sloc]) #next tile to parse
    tile_wires = defaultdict(list)
    tile_wires[sloc].append(snode) #wires in this tile we are interested in
    seen = set()


    while next_tile:
        tile_loc = next_tile.pop()
        srow, scol = tile_loc
        tile_bels = db[srow, scol].bels

        if tile_loc not in tile_dict or not tile_wires[tile_loc]:
            continue

        _, all_tile_wires, _ = parse_tile_(db, srow, scol, tile_dict[(srow,scol)])
        wire_dict = defaultdict(list)
        for dest_wire, src_wire in all_tile_wires.items():
            wire_dict[src_wire].append(dest_wire)

        # Trace the input of an IOB to its output
        for iob in ("IOBA", "IOBB"):
            if iob in tile_bels:
                portmap = tile_bels[iob].portmap
                input_node, output_node = portmap["I"], portmap["O"]
                wire_dict[input_node].append(output_node)

        wire_deque = deque(tile_wires[tile_loc]) #copy wires of interest
        tile_wires[tile_loc].clear() # reset the list of wires we are interested in

        while wire_deque:
            curr_wire = wire_deque.pop()
            full_node_id = (*tile_loc, curr_wire)

            # General Commment: It's possible to have loops e.g LUT -> FF -> LUT
            if full_node_id in seen:
                continue
            seen.add(full_node_id)

            #Trace through LUTs by adding the lut_output to the deque of wires we'll trace through
            if 'lut' in through_bel and re.match(LUT_IN_REGEX, curr_wire) and f"LUT{curr_wire[-1]}" in tile_bels:
                lut_output = f'F{curr_wire[-1]}'
                lut_output_id = (*tile_loc, lut_output)
                prior_input_id =  path_dict.get(lut_output_id)
                if prior_input_id:
                    lut_input_id = (*tile_loc, prior_input_id[2] + "#" + curr_wire)
                else:
                    lut_input_id = (*tile_loc, curr_wire)
                    wire_deque.append(lut_output)
                path_dict[lut_output_id] = lut_input_id

            #Trace through LUTs by adding the lut_output to the deque of wires we'll trace through
            if 'ff' in through_bel and re.match(FF_IN_REGEX, curr_wire) and f"DFF{curr_wire[-1]}" in tile_bels:
                ff_output = f'Q{dest_wire[-1]}'
                path_dict[(*tile_loc, ff_output)] = full_node_id
                wire_deque.append(ff_output)

            #Trace through intertile wires. No support for other aliases yet
            for itrow, itcol, inter_wire in next_intertile_wires(db,full_node_id):
                loc = (itrow, itcol)
                next_tile.append(loc)
                tile_wires[loc].append(inter_wire)
                path_dict[(*loc,inter_wire)] = full_node_id

            # Regular search stuff
            for dest_wire in wire_dict[curr_wire]:
                wire_id = (*tile_loc, dest_wire)
                wire_deque.appendleft(dest_wire)
                path_dict[wire_id] = full_node_id

    return path_dict



def get_output_path_dict(tile_dict:dict, db:Device, source:Node|str, through_bel=('lut', 'ff')) -> dict:
    """
    Returns dictionary of connections that lead to a `node`.
    Args:
        tile_dict (dict): Dictionary of tiles (output of tile_bitmap) .
        db (Device): Device object representing the FPGA or chip.
        source (Node|str): Node to start trace from. A bit of a misnomer since we are tracing from destination to source nodes
        through_bel (tuple, optional): BEL types to trace through. Defaults to ('lut', 'ff'), the only ones implemented for now.

    Returns:
        path_dict: A dictionary of dest<-src connections where the keys are destinations.
              In cases where there are multiple conceptual inputs to the same destination (like tracing from the output to input of a bel),
              the input wires are separated by a "#". dest and src are always strings
    """

    source = __normalize_pin(db, source)
    #parent of source is destination
    path_dict = {source: source}
    srow, scol, source_node = source
    sloc = (srow, scol)
    source_tiles = deque([sloc])
    tile_wires = defaultdict(list)
    tile_wires[sloc].append(source_node)

    LUT_OUT_REGEX = r'F([0-7])'
    FF_OUT_REGEX = r'Q([0-7])'

    seen = set()

    while source_tiles:
        tile_loc = source_tiles.pop()
        if tile_loc not in tile_dict or not tile_wires[tile_loc]:
            continue

        wire_deque = deque(tile_wires[tile_loc])
        tile_wires[tile_loc].clear()
        _, wire_dict, _ = parse_tile_(db, *tile_loc, tile_dict[tile_loc], default=True)

        # Trace the output of an IOB to its input
        tile_bels = db[tile_loc[0], tile_loc[1]].bels
        for iob in ("IOBA", "IOBB"):
            if iob in tile_bels:
                portmap = tile_bels[iob].portmap
                input_node, output_node = portmap["I"], portmap["O"]
                wire_dict[output_node] = input_node

        while wire_deque:
            curr_wire = wire_deque.pop()
            full_curr_id = (*tile_loc, curr_wire)
            if full_curr_id in seen:
                continue
            seen.add(full_curr_id)
            next_wire = wire_dict.get(curr_wire)

            if next_wire:
                wire_deque.append(next_wire)
                path_dict[full_curr_id] = (*tile_loc, next_wire)

            # Trace LUTs and FFs to their possible inputs.
            if 'lut' in through_bel and re.match(LUT_OUT_REGEX, curr_wire) and f"LUT{curr_wire[-1]}" in tile_bels:
                lut_sources = [F'{id}{curr_wire[-1]}' for id in ('A','B','C','D')]
                lut_sources = [x for x in lut_sources if x in wire_dict]
                if lut_sources:
                    #Assumption that there is never a '#' in a wirename, which holds so far.
                    path_dict[full_curr_id] = (*tile_loc, "#".join(lut_sources))
                    wire_deque.extend(lut_sources)

            if 'ff' in through_bel and re.match(FF_OUT_REGEX, curr_wire) and f"DFF{curr_wire[-1]}" in tile_bels:
                if (ff_source := f'F{curr_wire[-1]}') in wire_dict:
                    wire_deque.append(ff_source)
                    path_dict[full_curr_id] = (*tile_loc, ff_source)

            #Trace an intertile wire to its source in a potentially different tile
            if is_intertile_node(curr_wire):
                it_row, it_col, it_wire = source_intertile_wire(db, full_curr_id)
                it_node = (it_row, it_col, it_wire)
                path_dict[full_curr_id] = it_node
                if (it_node) not in seen:
                    source_tiles.append((it_row,it_col))
                    tile_wires[(it_row, it_col)].append(it_wire)

    return path_dict

def get_path_dict(tile_dict:dict, db:Device, source:Node|str, through_bel=('lut', 'ff')) -> dict:
    """
    Returns dictionary of connections that lead to a `node` or emanate from it
    Args:
        tile_dict (dict): Dictionary of tiles (output of tile_bitmap) .
        db (Device): Device object representing the FPGA or chip.
        source (Node|str): Node to start trace from. A bit of a misnomer since we are tracing from destination to source nodes
        through_bel (tuple, optional): BEL types to trace through. Defaults to ('lut', 'ff'), the only ones implemented for now.

    Returns:
        path_dict: A dictionary of dest<-src connections where the keys are destinations.
              In cases where there are multiple conceptual inputs to the same destination (like tracing from the output to input of a bel),
              the input wires are separated by a "#". dest and src are always strings
    """
    input_path_dict = get_input_path_dict(tile_dict, db, source, through_bel)
    output_path_dict = get_output_path_dict(tile_dict, db, source, through_bel)
    path_dict = {**input_path_dict, **output_path_dict}
    return path_dict


def io_node_to_tbrl(db, node:Node):
    row, col, wire = node
    for bel_type, bel in db[row, col].bels.items():
        if bel_type.startswith("IOB") and wire in bel.portmap.values():
            tbrl_name = rc2tbrl_0(db, row, col, bel_type[-1])
            return tbrl_name

def get_io_nodes(db:Device):
    rows, cols = db.rows, db.cols
    nodes = []
    for row in range(rows):
        for col in (0, cols-1):
            bels = db[row, col].bels
            for bel_type, bel in bels.items():
                if bel_type.startswith("IOB"):
                    bel_input = bel.portmap["I"]
                    nodes.append ((row, col, bel_input))

    for col in range(cols):
        for row in (0, rows-1):
            bels = db[row, col].bels
            for bel_type, bel in bels.items():
                if bel_type.startswith("IOB"):
                    bel_input = bel.portmap["I"]
                    nodes.append ((row, col, bel_input))

    return nodes

# Connected by default in tiles (src -> list[destinations]). No longer in tiledata from V1.9.9
default_pips = {
    "F0": ["D2", "D3", "D4", "D5", "D6", "D7"],
    "F1": ["B2", "B3", "B4", "B5", "B6", "B7"],
    "F2": ["D0", "D1"],
    "F3": ["B0", "B1"],
    "F4": ["C0", "C1", "C2", "C3", "C6", "C7"],
    "F5": ["A0", "A1", "A2", "A3", "A6", "A7"],
    "F6": ["C4", "C5"],
    "F7": ["A4", "A5"],
    "Q0": ["X01", "X05", "E100", "W100", "W130", "E130", "N100", "S100", "S130", "N130", "E200", "W200", "N200", "S200", "E800", "W800", "N800",  "S800"],
    "Q1": ["X02", "X06", "E210", "W210", "N210", "S210", "E810", "W810", "N810", "S810"],
    "Q2": ["E220", "W220", "N220", "S220"],
    "Q3": ["E230", "W230", "N230", "S230"],
    "Q4": ["E240", "W240", "N240", "S240", "E820", "W820", "N820", "S820"],
    "Q5": ["E250", "W250", "N250", "S250", "E830", "W830", "N830", "S830"],
    "Q6": ["X03", "X07", "E260", "W260", "N260", "S260"],
    "Q7": ["X04", "X08", "E270", "W270", "N270", "S270"]
}


# ratio of wires that are connected by default. I was getting multiple routes between the same
# source and destination at some point and the routes with more non-default wires seemed more
# plausible.
def assess_path(path:list[Node]):
    """
    Calculate the ratio of wires connected by default in a given path.

    Args:
        path (list[Node]): A list of nodes representing a route.

    Returns:
        float: The ratio of default-connected wires to total wires in the path.
    """
    qual = 0

    if len(path) < 2:
        return qual

    for i in range(len(path)-1):
        (srow, scol, snode), (drow, dcol, dnode) =  path[i], path[i+1]
        if srow==drow and scol==dcol and dnode in default_pips.get(snode, {}):
            qual += 1

    return (len(path) - qual)/len(path)

def enumerate_paths(path_dict:dict, sources:set[Node|str], dests:set[Node|str])->Iterable[Node]:
    """
    Generate all paths between given sources and destinations. Enumerating might be a better option
    than getting the list in one go for large designs.

    Args:
        path_dict (dict[dst, src]): A dictionary mapping destination nodes to their sources.
        sources (set[Node | str]): A set of starting nodes.
        dests (set[Node | str]): A set of destination nodes.

    Returns:
        Iterable[Node]: An iterator yielding paths as sequences of nodes from src to destination.
    """

    sources = set(sources)
    dests = list(set(dests))
    dests.sort()

    def __path_finder (prior_path):
        curr_node = prior_path[-1]
        path = list(prior_path)
        while ((next_node:=path_dict.get(curr_node)) not in path) and next_node:
            curr_node = next_node

            row, col, wire = curr_node
            if curr_node in sources or wire in ['VSS', 'VCC']:
                path.append(curr_node)
                yield list(path[::-1])

            #Assumption that there is never a '#' in a wirename, which holds so far...
            elif "#" in wire:
                alts = wire.split("#")
                for alt in alts:
                    node = (row, col, alt)
                    alt_path = [*path, node]
                    if node in sources and node not in path:
                        yield list(alt_path[::-1])
                    yield from __path_finder(alt_path)

            else:
                path.append(curr_node)

    for dst in dests:
        # print ('dst', dst)
        # print('dst', dst)
        yield from __path_finder([dst])


# Accumulate paths from `enumerate_paths` and optionally rank them
def get_paths(path_dict:dict, sources:set[Node|str], dests:set[Node|str], sort:bool=False):
    """
    Accumulate paths between sources and destinations (generated by `enumerate_paths`), with optional ranking.

    Args:
        path_dict (dict[dst, src]): A dictionary mapping destination nodes to their sources.
        sources (set[Node | str]): A set of starting nodes.
        dests (set[Node | str]): A set of destination nodes.
        sort (bool): Whether to sort the paths based on a ranking criteria (default is False).

    Returns:
        list: A list of paths between the sources and destinations, optionally sorted.
    """
    paths = []
    for path in enumerate_paths(path_dict, sources, dests):
        paths.append(path)

    if sort:
        paths.sort(key = lambda x: (x[0], x[-1], assess_path(x)))

    return paths

def visualize_grid(plot:dict[str, list[Node]], rows, cols, show=False, save_name='', checkbox=False):
    """
    Visualize a grid with marked tiles and annotations. The source tile for a path is marked with '^',
    while its destination tile is marked with '$'

    Parameters:
        plot (dict): A dictionary where keys represent group labels and values are lists of `Node` objects.
                     Each group of nodes will be marked with a distinct ASCII character.
                     Note: Avoid using `[all]` as a key in the plot dictionary.
        rows (int, optional): The number of rows in the grid. If not provided, it will be inferred from the data.
        cols (int, optional): The number of columns in the grid. If not provided, it will be inferred from the data.
        show (bool, optional): If `True`, the plot will be displayed. Defaults to `False`.
        save_name (str, optional): The filename to save the plot. If not provided, the plot will not be saved.
        checkbox (bool, optional): If `True`, adds interactive checkboxes to toggle the visibility of plot groups.
    Returns:
        None
    """
    try:
        import matplotlib
        import matplotlib.pyplot as plt
    except:
        raise ModuleNotFoundError ("Kindly install `matplotlib` to call this function")

    if checkbox:
        # matplotlib.use('Qt5Agg')
        from matplotlib.widgets import CheckButtons

    grid = [[0] * cols for _ in range(rows+1)]
    fig, ax = plt.subplots(figsize=(16,14))

    cmap = plt.get_cmap('tab10', 1)
    plt.imshow(grid, cmap=cmap, origin='upper', extent=[0, cols, 0, rows])
    plt.tick_params(labeltop=True, labelright=True, labelsize=10)
    plt.grid(which='both', color='black', linestyle='-', linewidth=1)

    x_ticks = list(range(0, cols+1))
    plt.xticks(x_ticks, rotation=90, fontsize=8, fontweight='light', fontfamily='monospace')

    yticks = list(range(rows, -1, -1 ))
    ylabels = [str(x) for x in yticks[::-1]]
    plt.yticks(yticks, ylabels, fontsize=8, fontweight='light', fontfamily='monospace')

    # ax = plt.gca()
    ax.set_aspect(0.8, adjustable='box')
    # ax.set_adjustable()

    plt.subplots_adjust(left=0.1, right=0.6, wspace=0.05)
    plot_chars = []
    plot_chars.extend(list(string.ascii_uppercase))
    # colors = ["white", "yellow", "red", "black", "gold", "cyan", "orange"]
    colors = ["white", "black", "red", "gold"]
    legend_opts = list(itertools.product(plot_chars, colors))
    random.Random(12).shuffle(legend_opts)

    if isinstance(plot, list):
        plot = {0: plot}


    plot = {str(k):v for k, v in plot.items()} #Cast to Strings for consistency

    legend_dict = {}
    l_idx = 0
    for k, path in plot.items():
        legend = legend_opts[l_idx%len(legend_opts)]
        legend_dict[str(k)] = [*legend, f"({legend[0]}) {path[0]} -> {path[-1]} //{k}"] #plot_character, plot_color, plot_label
        l_idx += 1

    legend_entries = [
        plt.Line2D([0], [0], color=color, markerfacecolor=color, marker='o', linestyle='',
                   markeredgecolor='black', markersize=10, label=path_label)
        for k, (plot_char, color, path_label) in legend_dict.items()
    ]

    text_dict = defaultdict(list)
    for k, path in plot.items():
        base_plot_char, color, label = legend_dict[str(k)]

        locs = [(node[0], node[1]) for node in path]
        start_loc, end_loc = locs[0], locs[-1]

        for loc in locs:
            plot_char = base_plot_char
            if len(path) == 1:
                plot_char = plot_char + "!" #Single tile marker
            elif loc == start_loc:
                plot_char = plot_char + "^" #Start tile marker
            elif loc == end_loc:
                plot_char = plot_char + "$" #Final tile marker

            i, j = loc
            this_text = ax.text(j + 0.5, rows - i - 0.5, plot_char, ha='center', va='center', color=color,
                                fontweight='bold', fontsize=8, visible=True)
            text_dict[k].append((loc, this_text)) # We store the texts so we can easily toggle visibility later on

    ALL_TAG = "[ALL]"
    _showing = set(plot.keys())
    showing = set(_showing) #We'll use this to manage state

    def check_callback(to_toggle=''):
        # print('click registered', "to_toggle", to_toggle)
        check.eventson = False
        if to_toggle is None:
            return

        if to_toggle != ALL_TAG:
            _, key = to_toggle.split("//")
            text_objs = text_dict[key]
            visibility = key in showing
            for loc, text in text_objs:
                text.set_visible(not visibility)
            if visibility:
                showing.remove(key)
            else:
                showing.add(key)

        else:
            key = ALL_TAG
            visibility = len(showing) == len(_showing)
            for text_list in text_dict.values():
                for text_loc, text_obj in text_list:
                    text_obj.set_visible(not visibility) #Toggle visibiility of characters

            if visibility:
                check.clear() #Clear all checkboxes
                showing.clear()
            else:
                for i in range(len(_showing)):
                    check.set_active(i, True) #Activate all checkboxes
                showing.update(_showing)

        # Keep the state of the 'ALL' checkbox consistent
        if len(showing) < len(_showing):
            check.set_active(0, False)
        elif len(showing) == len(_showing):
            check.set_active(0, True)

        check.eventson = True
        plt.draw()

    # Labels for checkboxes
    check_labels = [ALL_TAG]
    check_labels.extend([v[2] for k, v in legend_dict.items()])

    if show and checkbox:
        rax = plt.axes([0.6, 0.1, 0.4, 0.8])  # Position [left, bottom, width, height]
        rax.autoscale()
        rax.set_frame_on(False)
        check = CheckButtons(rax, check_labels, [True] * len(check_labels))
        check.on_clicked(check_callback)
    elif not (show or checkbox):
        plt.legend(
            handles=legend_entries,
            loc='center left',  # Legend location
            bbox_to_anchor=(1.05, 0.5),  # Position outside the grid
            fontsize=10
        )

    if save_name:
        plt.savefig(save_name+".jpeg", dpi=300, bbox_inches='tight')

    if show:
        plt.axes(ax)
        plt.show()



import pypdf
import re
from pathlib import Path
from typing import List, Dict
import ollama
import os
import json
import glob
from collections import defaultdict
from pypdf import PdfReader

def get_outline_tree(reader: PdfReader) -> List[Dict]:
    """
    Extract the complete outline/bookmark structure from a PDF.
    
    Args:
        reader (PdfReader): PyPDF Reader object
        
    Returns:
        list: List of dictionaries containing outline information
    """
    outlines = reader.outline
    
    if not outlines:
        return []

    def extract_destinations(outline_item):
        """Recursively extract all destinations from an outline item"""
        if isinstance(outline_item, list):
            return [item for i in outline_item for item in extract_destinations(i)]
            
        if not hasattr(outline_item, '__getitem__'):
            return []
            
        try:
            title = outline_item.get('/Title', 'Untitled')
            page_num = reader.get_destination_page_number(outline_item)
            
            result = [{
                'title': title,
                'page': page_num
            }]
            
            # Check for children (/First and /Next form a linked list of children)
            if '/First' in outline_item:
                result.extend(extract_destinations(outline_item['/First']))
            
            # Check for siblings
            if '/Next' in outline_item:
                result.extend(extract_destinations(outline_item['/Next']))
                
            return result
            
        except Exception as e:
            print(f"Error processing outline item: {e}")
            return []

    return extract_destinations(outlines)

def find_sections_in_outline(outline: List[Dict], pattern: str, flags: int = re.IGNORECASE) -> List[Dict]:
    """
    Search for sections in the outline that match the regex pattern.
    
    Args:
        outline (list): Outline structure from get_outline_tree
        pattern (str): Regular expression pattern to search for
        flags (int): Regular expression flags (default: case insensitive)
        
    Returns:
        list: Matching sections with their details
    """
    matches = []
    
    try:
        regex = re.compile(pattern, flags)
        
        for item in outline:
            title = item.get('title', '')
            if regex.search(title):
                matches.append(item)
    
    except re.error as e:
        print(f"Invalid regular expression: {e}")
        return []
        
    return matches

def print_outline(outline: List[Dict]) -> None:
    """
    Print the outline structure.
    
    Args:
        outline (list): Outline structure from get_outline_tree
    """
    if not outline:
        return
        
    for item in outline:
        title = item.get('title', 'Untitled')
        page = item.get('page', 0)
        print(f"- {title} (Page {page + 1})")


def analyze_primitives(json_pattern):
    primitive_counts = defaultdict(int)
    
    for filename in glob.glob(json_pattern):
        with open(filename) as f:
            design = json.load(f)
            
        for module in design['modules'].values():
            for cell in module.get('cells', {}).values():
                primitive_counts[cell['type']] += 1
    
    # Sort by count descending
    return dict(sorted(primitive_counts.items(), key=lambda x: x[1], reverse=True))


def generate_summary(pdfs, primitive_name: str) -> str:
    """Generate markdown summary using Ollama."""

    doc_sections = []
    for reader in pdfs:
        outline = get_outline_tree(reader)
        matches = find_sections_in_outline(outline, rf"^[0-9\.]+ +{primitive_name}$")
        if matches:
            pages = [reader.pages[match['page']].extract_text() for match in matches]
            doc_sections.append(pages)

    doc_context = "\n\n".join(
        "\n".join(pages)
        for pages in doc_sections
    )
    
    description = ""
    if doc_context:
        prompt = f"""
{doc_context}

Summarizing the functionality of the Gowin {primitive_name} primitive.
Stick to the provided facts and do not make assumptions.
Write only the requested paragraph and nothing else.
"""
        print(prompt)
        response = ollama.generate(
            model='llama3.1:8b',
            prompt=prompt,
        )
        description = escape_markdown(response["response"])
        print(description)
    return description


def escape_markdown(text):
    return text.replace('_', '\\_')

def format_bits(bits):
    if isinstance(bits, list):
        if len(bits) == 1:
            return "1"
        else:
            return str(len(bits))
    return "1"

def get_default_param_value(value):
    if isinstance(value, str) and all(c in '01' for c in value):
        decimal_value = int(value, 2)
        return f"{decimal_value} (0b{value})"
    return value

def generate_module_doc(module_name, module_data, supported, pdfs):
    module_doc = []
    
    module_doc.append(generate_summary(pdfs, module_name))
    
    module_doc.append("\n")
    if supported:
        module_doc.append("This device is supported in Apicula.")
    else:
        module_doc.append("This device is not yet supported in Apicula")

    # Ports table
    if module_data.get('ports'):
        module_doc.append("## Ports\n")
        module_doc.append("| Port | Size | Direction |")
        module_doc.append("|------|------|-----------|")
        
        for port_name, port_data in sorted(module_data['ports'].items()):
            size = format_bits(port_data['bits'])
            direction = port_data['direction']
            module_doc.append(f"| {escape_markdown(port_name)} | {size} | {direction} |")
        module_doc.append("\n")
    
    # Parameters table
    if module_data.get('parameter_default_values'):
        module_doc.append("## Parameters\n")
        module_doc.append("| Parameter | Default Value |")
        module_doc.append("|-----------|---------------|")
        
        for param_name, param_value in sorted(module_data['parameter_default_values'].items()):
            default_value = get_default_param_value(param_value)
            module_doc.append(f"| {escape_markdown(param_name)} | {default_value} |")
        module_doc.append("\n")
    
    # Verilog instantiation template
    module_doc.append("## Verilog Instantiation")
    module_doc.append("```verilog")
    
    # Generate instantiation
    instance = []
    
    # Add parameters if they exist
    if module_data.get('parameter_default_values'):
        instance.append(f"{module_name} #(")
        param_list = []
        for param_name in sorted(module_data['parameter_default_values'].keys()):
            param_list.append(f"    .{param_name}({param_name})")
        instance.append(",\n".join(param_list))
        instance.append(f") {module_name.lower()}_inst (")
    else:
        instance.append(f"{module_name} {module_name.lower()}_inst (")
    
    # Add ports
    if module_data.get('ports'):
        port_list = []
        for port_name in sorted(module_data['ports'].keys()):
            port_list.append(f"    .{port_name}({port_name})")
        instance.append(",\n".join(port_list))
    
    instance.append(");")
    module_doc.append("\n".join(instance))
    module_doc.append("```\n")
    
    return "\n".join(module_doc)

def process_json_file(json_data, stats, pdfs, output_dir):
    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)
    
    # Process each module
    for module_name, module_data in sorted(json_data['modules'].items()):
        # Generate documentation for this module
        supported = module_name in stats
        doc = generate_module_doc(module_name, module_data, supported, pdfs)
        
        # Write to file
        filename = os.path.join(output_dir, f"{module_name}.md")
        with open(filename, 'w') as f:
            f.write(doc)

def main():
    # Create docs directory
    output_dir = "../apicula.wiki"
    
    # Read and process the JSON file
    with open('prims.json', 'r') as f:
        data = json.load(f)
    
    stats = analyze_primitives('examples/himbaechel/*-synth.json')

    pdfs = []
    for pdf_path in Path("~/Documents/gowin/").expanduser().glob("*.pdf"):
        pdfs.append(pypdf.PdfReader(pdf_path))
    
    # Generate documentation
    process_json_file(data, stats, pdfs, output_dir)
    print(f"Documentation generated in {output_dir}/")

if __name__ == "__main__":
    main()


# pll tool to find best match for the target frequency
# calculations based on: https://github.com/juj/gowin_fpga_code_generators/blob/main/pll_calculator.html
# limits from:
# - http://cdn.gowinsemi.com.cn/DS117E.pdf,
# - http://cdn.gowinsemi.com.cn/DS861E.pdf,
# - https://cdn.gowinsemi.com.cn/DS226E.pdf

import argparse
import re
import sys


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i", "--input-freq-mhz", help="PLL Input Frequency", type=float, default=27
    )
    parser.add_argument(
        "-o", "--output-freq-mhz", help="PLL Output Frequency", type=float, default=108
    )
    parser.add_argument(
        "-d", "--device", help="Device", type=str, default="GW1NR-9 C6/I5"
    )
    parser.add_argument(
        "-f",
        "--filename",
        help="Save PLL configuration as Verilog to file",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-m",
        "--module-name",
        help="Specify different Verilog module name than the default 'pll'",
        type=str,
        default="pll",
    )
    parser.add_argument("-l", "--list-devices", help="list device", action="store_true")

    args = parser.parse_args()

    device_name = args.device
    match = re.search(
        r"(GW[125][A-Z]{1,3})-[A-Z]{0,2}([0-9]{1,2})[A-Z]{1,3}[0-9]{1,3}P*N*(C[0-9]/I[0-9]|ES)",
        device_name,
    )
    if match:
        device_name = f"{match.group(1)}-{match.group(2)} {match.group(3)}"
    else:
        print(f'Warning: cannot decipher the name of the device {device_name}.')

    device_limits = {
        "GW1N-1 C6/I5": {
            "comment": "Untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 900,
            "clkout_min": 3.125,
            "clkout_max": 450,
        },
        "GW1N-1 C5/I4": {
            "comment": "Untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 320,
            "vco_max": 720,
            "clkout_min": 2.5,
            "clkout_max": 360,
        },
        "GW1NR-2 C7/I6": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 800,
            "clkout_min": 3.125,
            "clkout_max": 750,
        },
        "GW1NR-2 C6/I5": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 800,
            "clkout_min": 3.125,
            "clkout_max": 750,
        },
        "GW1NR-2 C5/I4": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 320,
            "vco_max": 640,
            "clkout_min": 2.5,
            "clkout_max": 640,
        },
        "GW1NR-4 C6/I5": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1000,
            "clkout_min": 3.125,
            "clkout_max": 500,
        },
        "GW1NR-4 C5/I4": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 320,
            "vco_max": 800,
            "clkout_min": 2.5,
            "clkout_max": 400,
        },
        "GW1NSR-4 C7/I6": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NSR-4 C6/I5": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NSR-4 C5/I4": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 320,
            "vco_max": 960,
            "clkout_min": 2.5,
            "clkout_max": 480,
        },
        "GW1NSR-4C C7/I6": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NSR-4C C6/I5": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NSR-4C C5/I4": {
            "comment": "Untested",
            "pll_name": "PLLVR",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 320,
            "vco_max": 960,
            "clkout_min": 2.5,
            "clkout_max": 480,
        },
        "GW1NR-9 C7/I6": {
            "comment": "Untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NR-9 C6/I5": {
            "comment": "tested on TangNano9K Board",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 1200,
            "clkout_min": 3.125,
            "clkout_max": 600,
        },
        "GW1NR-9 C6/I4": {
            "comment": "Untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 320,
            "vco_min": 3200,
            "vco_max": 960,
            "clkout_min": 2.5,
            "clkout_max": 480,
        },
        "GW1NZ-1 C6/I5": {
            "comment": "untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 400,
            "vco_min": 400,
            "vco_max": 800,
            "clkout_min": 3.125,
            "clkout_max": 400,
        },
        "GW2A-18 C8/I7": {
            "comment": "untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 500,
            "vco_min": 500,
            "vco_max": 1250,
            "clkout_min": 3.90625,
            "clkout_max": 625,
        },
        "GW2AR-18 C8/I7": {
            "comment": "untested",
            "pll_name": "rPLL",
            "pfd_min": 3,
            "pfd_max": 500,
            "vco_min": 500,
            "vco_max": 1250,
            "clkout_min": 3.90625,
            "clkout_max": 625,
        },
        "GW5A-25 ES": {
            "comment": "untested",
            "pll_name": "rPLL",
            "pfd_min": 19,
            "pfd_max": 800,
            "vco_min": 800,
            "vco_max": 1600,
            # The previous four parameters are taken from the datasheet (as in
            # this case from https://cdn.gowinsemi.com.cn/DS1103E.pdf), but I
            # don't know where these two come from:(
            "clkout_min": 6.25,
            "clkout_max": 1600,
        },
    }

    if args.list_devices:
        for device in device_limits:
            print(f"{device} - {device_limits[device]['comment']}")
        sys.exit(0)

    if device_name not in device_limits:
        print(f"ERROR: device '{device_name}' not found")
        sys.exit(1)

    limits = device_limits[device_name]
    setup = {}

    FCLKIN = args.input_freq_mhz
    min_diff = FCLKIN

    for IDIV_SEL in range(64):
        for FBDIV_SEL in range(64):
            for ODIV_SEL in [2, 4, 8, 16, 32, 48, 64, 80, 96, 112, 128]:
                PFD = FCLKIN / (IDIV_SEL + 1)
                if not (limits["pfd_min"] <= PFD <= limits["pfd_max"]):
                    continue
                CLKOUT = FCLKIN * (FBDIV_SEL + 1) / (IDIV_SEL + 1)
                if not (limits["clkout_min"] < CLKOUT < limits["clkout_max"]):
                    continue
                VCO = (FCLKIN * (FBDIV_SEL + 1) * ODIV_SEL) / (IDIV_SEL + 1)
                if not (limits["vco_min"] < VCO < limits["vco_max"]):
                    continue
                diff = abs(args.output_freq_mhz - CLKOUT)
                if diff < min_diff:
                    min_diff = diff
                    setup = {
                        "IDIV_SEL": IDIV_SEL,
                        "FBDIV_SEL": FBDIV_SEL,
                        "ODIV_SEL": ODIV_SEL,
                        "PFD": PFD,
                        "CLKOUT": CLKOUT,
                        "VCO": VCO,
                        "ERROR": diff,
                    }

    if setup:

        extra_options = ""
        if limits["pll_name"] == "PLLVR":
            extra_options = ".VREN(1'b1),"

        pll_v = f"""/**
 * PLL configuration
 *
 * This Verilog module was generated automatically
 * using the gowin-pll tool.
 * Use at your own risk.
 *
 * Target-Device:                {device_name}
 * Given input frequency:        {args.input_freq_mhz:0.3f} MHz
 * Requested output frequency:   {args.output_freq_mhz:0.3f} MHz
 * Achieved output frequency:    {setup['CLKOUT']:0.3f} MHz
 */

module {args.module_name}(
        input  clock_in,
        output clock_out,
        output locked
    );

    {limits['pll_name']} #(
        .FCLKIN("{args.input_freq_mhz}"),
        .IDIV_SEL({setup['IDIV_SEL']}), // -> PFD = {setup['PFD']} MHz (range: {limits['pfd_min']}-{limits['pfd_max']} MHz)
        .FBDIV_SEL({setup['FBDIV_SEL']}), // -> CLKOUT = {setup['CLKOUT']} MHz (range: {limits['clkout_min']}-{limits['clkout_max']} MHz)
        .ODIV_SEL({setup['ODIV_SEL']}) // -> VCO = {setup['VCO']} MHz (range: {limits['vco_min']}-{limits['vco_max']} MHz)
    ) pll (.CLKOUTP(), .CLKOUTD(), .CLKOUTD3(), .RESET(1'b0), .RESET_P(1'b0), .CLKFB(1'b0), .FBDSEL(6'b0), .IDSEL(6'b0), .ODSEL(6'b0), .PSDA(4'b0), .DUTYDA(4'b0), .FDLY(4'b0), {extra_options}
        .CLKIN(clock_in), // {args.input_freq_mhz} MHz
        .CLKOUT(clock_out), // {setup['CLKOUT']} MHz
        .LOCK(locked)
    );

endmodule
"""
        if args.filename:
            open(args.filename, "w").write(pll_v)
        else:
            print(pll_v)


if __name__ == "__main__":
    main()



import sys
import json
import numpy as np
from bslib import read_bitstream
from PIL import Image

image = np.zeros([712, 2840], dtype="byte")
for fname in sys.argv[1:]:
    print(fname)
    with open(fname) as f:
        try:
            data = json.load(f)
        except json.decoder.JSONDecodeError:
            continue
        for x, y in data:
            image[x][y] += 1

print(np.nonzero(image > 1))
im = Image.frombytes(mode='1', size=image.shape[::-1], data=np.packbits(image))
#im.show()
im.save("bitmap.png","PNG")



#!/usr/bin/python3

import os
import re
import subprocess

def run_script(port : str, pinName : str):
    with open('template.vhd', 'rt') as f:
        template = f.read()

    f.close()

    template = template.replace("##PORT##", port)
    template = template.replace("##PORTNAME##", pinName)
    with open('findpin.vhd', 'wt') as f2:
        f2.write(template)

    f2.close()

    GowinHome = os.getenv('GOWINHOME')

    result = subprocess.run(["rm", "unpack.v"])
    result = subprocess.run([GowinHome+"/IDE/bin/gw_sh", "findpin.tcl"], check=True)
    result = subprocess.run(["gowin_unpack", "-d", "GW1N-9", "impl/pnr/project.fs"], check=True)

    with open('unpack.v', 'rt') as f3:
        unpackv = f3.readline()

    iobre = re.compile("\(([a-zA-Z0-9_]*)\)")

    matches = iobre.findall(unpackv)

    print("Found IOB: ", matches[0])

    with open('pin_report.txt', 'a+') as f:
        f.write(pinName + " -> " + matches[0] + "\n")

    f.close()


######################################################
## MAIN                                             ##
######################################################

result = subprocess.run(["rm", "pin_report.txt"])

# pin name, bus width, bus offset
pins = [
    ["IO_sdram_dq", 16, 0],
    ["O_sdram_clk", 0, 0],
    ["O_sdram_cke", 0, 0],
    ["O_sdram_cs_n", 0, 0],
    ["O_sdram_cas_n", 0, 0],
    ["O_sdram_ras_n", 0, 0],
    ["O_sdram_wen_n", 0, 0],
    ["O_sdram_addr", 12, 0],
    ["O_sdram_dqm", 2, 0],
    ["O_sdram_ba", 2, 0]
]



for pin in pins:
    pinName = pin[0]
    pinBuswidth = pin[1]
    pinOffset = pin[2]

    if (pinBuswidth == 0):
        port = pinName
        if (pinName.startswith("O")):
            port = port + " : out std_logic"
        else:
            port = port + " : inout std_logic"

        run_script(port, pinName)
    else:
        for pinIdx in range(0, pinBuswidth):
            port = pinName
            if (pinName.startswith("O")):
                port = port + " : out "
            else:
                port = port + " : inout "

            actualIdx = pinIdx + pinOffset

            port = port + "std_logic_vector(" + str(actualIdx) + " downto " + str(actualIdx) + ")"

            run_script(port, pinName+"("+str(actualIdx)+")")



import sys
import os
from itertools import chain
import re
import code
import pickle
import importlib.resources
sys.path.append(os.path.join(sys.path[0], '..'))
from apycula import chipdb
from apycula.wirenames import wirenames

device = os.getenv("DEVICE")
if not device:
    raise Exception("DEVICE not set")

with importlib.resources.open_binary("apycula", f"{device}.pickle") as f:
    db = pickle.load(f)

timing_class = "C6/I5" # TODO parameterize
timing = db.timing[timing_class]

def addWire(row, col, wire):
    gname = chipdb.wire2global(row, col, db, wire)
    #print("wire", gname)
    try:
        ctx.addWire(name=gname, type=wire, y=row, x=col)
    except AssertionError:
        pass
        #print("duplicate wire")

belre = re.compile(r"(IOB|LUT|DFF|BANK|CFG)(\w*)")
for row, rowdata in enumerate(db.grid, 1):
    for col, tile in enumerate(rowdata, 1):
        # add wires
        wires = set(chain(
            tile.pips.keys(),
            tile.clock_pips.keys(),
            *tile.pips.values(),
            *tile.clock_pips.values(),
            ))
        for wire in wires:
            addWire(row, col, wire)
        # add aliasses
        # creat bels
        #print(row, col, ttyp)
        for name, bel in tile.bels.items():
            typ, idx = belre.match(name).groups()
            if typ == "DFF":
                z = int(idx)
                belname = f"R{row}C{col}_SLICE{z}"
                clkname = f"R{row}C{col}_CLK{z//2}"
                fname = f"R{row}C{col}_F{z}"
                qname = f"R{row}C{col}_Q{z}"
                #print("IOB", row, col, clkname, fname, qname)
                ctx.addBel(name=belname, type="GENERIC_SLICE", loc=Loc(col, row, z), gb=False)
                ctx.addBelInput(bel=belname, name="CLK", wire=clkname)
                for k, n in [('A', 'I[0]'), ('B', 'I[1]'),
                               ('C', 'I[2]'), ('D', 'I[3]')]:
                    inpname = f"R{row}C{col}_{k}{z}"
                    ctx.addBelInput(bel=belname, name=n, wire=inpname)
                ctx.addBelOutput(bel=belname, name="Q", wire=qname)
                ctx.addBelOutput(bel=belname, name="F", wire=fname)

            elif typ == "IOB":
                z = ord(idx)-ord('A')
                belname = f"R{row}C{col}_IOB{idx}"
                inp = bel.portmap['I']
                outp = bel.portmap['O']
                oe = bel.portmap['OE']
                iname = f"R{row}C{col}_{inp}"
                oname = f"R{row}C{col}_{outp}"
                oename = f"R{row}C{col}_{oe}"
                #print("IOB", row, col, iname, oname, oename)
                ctx.addBel(name=belname, type="GENERIC_IOB", loc=Loc(col, row, z), gb=False)
                ctx.addBelInput(bel=belname, name="I", wire=iname)
                ctx.addBelInput(bel=belname, name="EN", wire=oename)
                ctx.addBelOutput(bel=belname, name="O", wire=oname)


wlens = {0: 'X0', 1: 'FX1', 2: 'X2', 8: 'X8'}
def wiredelay(wire):
    m = re.match(r"[NESWX]+([0128])", wire)
    if m:
        wlen = int(m.groups()[0])
        name = wlens.get(wlen)
        return ctx.getDelayFromNS(max(timing['wire'][name]))
    elif wire.startswith("UNK"):
        return ctx.getDelayFromNS(max(timing['glbsrc']['PIO_CENT_PCLK']))
    elif wire.startswith("SPINE"):
        return ctx.getDelayFromNS(max(timing['glbsrc']['CENT_SPINE_PCLK']))
    elif wire.startswith("GT"):
        return ctx.getDelayFromNS(max(timing['glbsrc']['SPINE_TAP_PCLK']))
    elif wire.startswith("GBO"):
        return ctx.getDelayFromNS(max(timing['glbsrc']['TAP_BRANCH_PCLK']))
    elif wire.startswith("GB"):
        return ctx.getDelayFromNS(max(timing['glbsrc']['BRANCH_PCLK']))
    else: # no known delay
        return ctx.getDelayFromNS(0)



def addPip(row, col, srcname, destname):
    gsrcname = chipdb.wire2global(row, col, db, srcname)
    gdestname = chipdb.wire2global(row, col, db, destname)

    pipname = f"R{row}C{col}_{srcname}_{destname}"
    #print("pip", pipname, srcname, gsrcname, destname, gdestname)
    try:
        # delay is crude fudge from vendor critical path
        ctx.addPip(
            name=pipname, type=destname, srcWire=gsrcname, dstWire=gdestname,
            delay=wiredelay(destname), loc=Loc(col, row, 0))
    except IndexError:
        pass
        #print("Wire not found", gsrcname, gdestname)
    except AssertionError:
        pass
        #print("Wire already exists", gsrcname, gdestname)

def addAlias(row, col, srcname, destname):
    gsrcname = chipdb.wire2global(row, col, db, srcname)
    gdestname = chipdb.wire2global(row, col, db, destname)

    pipname = f"R{row}C{col}_{srcname}_{destname}"
    #print("alias", pipname)
    ctx.addPip(
        name=pipname+"_ALIAS", type=destname, srcWire=gsrcname, dstWire=gdestname,
        delay=ctx.getDelayFromNS(0), loc=Loc(col, row, 0))


for row, rowdata in enumerate(db.grid, 1):
    for col, tile in enumerate(rowdata, 1):
        for dest, srcs in tile.pips.items():
            for src in srcs.keys():
                addPip(row, col, src, dest)
        for dest, srcs in tile.clock_pips.items():
            for src in srcs.keys():
                addPip(row, col, src, dest)
        for dest, src in tile.aliases.items():
            addAlias(row, col, src, dest)

for (row, col, pip), (srow, scol, spip) in db.aliases.items():
    dest = f"R{row+1}C{col+1}_{pip}"
    src = f"R{srow+1}C{scol+1}_{spip}"
    name = f"{src}_{dest}_ALIAS"
    ctx.addPip(
        name=name, type=dest, srcWire=src, dstWire=dest,
        delay=ctx.getDelayFromNS(0), loc=Loc(row+1, col+1, 0))

# too low numbers will result in slow routing iterations
# too high numbers will result in more iterations
ctx.setDelayScaling(0.1, 0.7)

#code.interact(local=locals())
